/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent.util.eventstream;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.json.JSONException;
import org.json.JSONObject;
import org.onosproject.restconf.api.RestconfRpcOutput;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.util.AbstractMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

import static org.onosproject.fpcagent.util.FpcUtil.restconfService;

/**
 * Implements a Worker to process event-data pairs sent by the FPC Client.
 */
public class NbEventWorkerManager implements AutoCloseable {
    private static final Logger log = LoggerFactory.getLogger(NbEventWorkerManager.class);
    private static NbEventWorkerManager _instance = null;

    public static final Map<Integer, String> CLIENT_ID_TO_URI;

    private final BlockingQueue<Entry<String, Entry<String, String>>> blockingQueue;
    private final int poolSize;
    private boolean run;

    static {
        CLIENT_ID_TO_URI = new ConcurrentHashMap<>();
    }

    protected NbEventWorkerManager(int poolSize) {
        this.poolSize = poolSize;
        this.blockingQueue = new LinkedBlockingQueue<>();
    }

    public static NbEventWorkerManager createInstance(int poolSize) {
        if (_instance == null) {
            _instance = new NbEventWorkerManager(poolSize);
        }
        return _instance;
    }

    public static NbEventWorkerManager getInstance() {
        return _instance;
    }

    public void send(Entry<String, Entry<String, String>> buf) {
        try {
            blockingQueue.put(buf);
        } catch (InterruptedException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    public void open() {
        Executors.newFixedThreadPool(this.poolSize).submit(() -> {
            this.run = true;
            Entry<String, Entry<String, String>> contents = null;
            while (run) {
                try {
                    contents = blockingQueue.take();

                    // call RESTCONF service
                    ObjectMapper mapper = new ObjectMapper();
                    log.debug("NB Event Contents: {}", contents.getValue().getValue());
                    ObjectNode node = (ObjectNode) mapper.readTree(contents.getValue().getValue());
                    if (contents.getValue().getKey().contains("ietf-dmm-fpcagent:configure")) {
                        CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                                new URI("/onos/restconf/operations/ietf-dmm-fpcagent:configure"),
                                node,
                                contents.getKey()
                        );
                        String replace = output.get().output().toString()
                                .replace("ietf-dmm-fpcagent:output", "fpc:output");
                        ConfigureService.responseQueue.add(
                                new AbstractMap.SimpleEntry(
                                        contents.getKey(),
                                        "event:application/json;/onos/restconf/" +
                                                "operations/ietf-dmm-fpcagent:configure\ndata:" + replace + "\r\n"
                                )
                        );
                    } else if (contents.getValue().getKey().contains("fpc:register-client")) {
                        CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                                new URI("/onos/restconf/operations/fpc:register-client"),
                                node,
                                contents.getKey()
                        );
                        addClientIdToUriMapping(output.get().output().toString(), contents.getKey());
                        ConfigureService.responseQueue.add(
                                new AbstractMap.SimpleEntry(
                                        contents.getKey(),
                                        "event:application/json;/onos/restconf/operations/fpc:register-client\ndata:" +
                                                output.get().output().toString() + "\r\n"
                                )
                        );
                    } else if (contents.getValue().getKey().contains("fpc:deregister-client")) {
                        CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                                new URI("/onos/restconf/operations/fpc:deregister-client"),
                                node,
                                contents.getKey()
                        );
                        removeClientIdToUriMapping(contents.getValue().getValue());
                    }

                } catch (Exception e) {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                }
            }
        });
    }

    /**
     * Add a FPC client URI to FPC client id mapping.
     *
     * @param registerClientOutput - Output of the register_cleint rpc
     * @param uri                  - FPC Client URI
     */
    private void addClientIdToUriMapping(String registerClientOutput, String uri) {
        try {
            JSONObject registerjson = new JSONObject(registerClientOutput);
//            log.info("registerjson: {}", registerjson);
            CLIENT_ID_TO_URI.put(
                    Integer.parseInt(registerjson.getJSONObject("fpc:output").getString("client-id")), uri
            );
        } catch (JSONException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    /**
     * Remove a FPC client URI to FPCClient Id mapping.
     *
     * @param deregisterClientInput - Input String of deregister_client rpc
     */
    private void removeClientIdToUriMapping(String deregisterClientInput) {
        try {
            JSONObject deregisterjson = new JSONObject(deregisterClientInput);
//            log.info("deregisterjson: {}", deregisterjson);
            CLIENT_ID_TO_URI.remove(
                    Integer.parseInt(deregisterjson.getJSONObject("input").getString("client-id"))
            );
        } catch (JSONException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    @Override
    public void close() {
        this.run = false;
    }
}

